"""a Python software that gets basic information about an MCPE server

RakLib Class

Copyright (c) 2016 w-gao
"""


class RakLib:

    PROTOCOL = 6
    MAGIC = b"\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x12\x34\x56\x78"

    UNCONNECTED_PING = 0x01
    UNCONNECTED_PONG = 0x1c



"""a Python software that gets basic information about an MCPE server

Packet class

Copyright (c) 2016 w-gao
"""

import struct


class Packet:
    def __init__(self):
        self.offset = 0
        self.buffer = b''

    def encode(self):
        pass

    def decode(self):
        self.offset = 0

    def read(self, length):

        if length < 0:
            self.offset = len(self.buffer) - 1
            return bytes(0)

        buff = bytearray()
        for i in range(length):
            self.offset += 1
            buff.append(self.buffer[self.offset])
        return buff

    def read_byte(self):
        self.offset += 1
        return self.buffer[self.offset]

    def read_short(self):
        return struct.unpack(">H", self.read(2))[0]

    def read_triad(self):
        return struct.unpack('>i', self.read(3) + b'\x00')[0]

    def read_int(self):
        return struct.unpack(">i", self.read(4))[0]

    def read_long(self):
        return struct.unpack(">q", self.read(8))[0]

    def read_string(self):
        l = self.read_short()
        return self.read(l).decode('utf-8')

    def write(self, val):
        self.buffer += val

    def write_byte(self, val):
        self.buffer += val

    def write_short(self, val):
        self.buffer += struct.pack('>h', val)

    def write_triad(self, val):
        self.buffer += struct.pack('>i', val)[:3]

    def write_int(self, val):
        self.buffer += struct.pack('>i', val)

    def write_long(self, val):
        self.buffer += struct.pack('>q', val)

    def write_string(self, val):
        b = val.encode('utf-8')
        self.write_short(len(b))
        self.buffer += b



"""a Python software that gets basic information about an MCPE server

UnconnectedPing class

Copyright (c) 2016 w-gao
"""

from .packet import Packet
from ..rak_lib import RakLib


class UnconnectedPing(Packet):

    def __init__(self, ping_id):
        super().__init__()

        self.ping_id = ping_id

    def encode(self):
        self.write_byte(bytes([RakLib.UNCONNECTED_PING]))
        self.write_long(self.ping_id)
        self.write(RakLib.MAGIC)

        # Guid
        self.write_long(0)

    def decode(self):
        pass



"""a Python software that gets basic information about an MCPE server

This package contains one sub-package and one module:

sub-packages:
packet

modules:
rak_lib

Copyright (c) 2016 w-gao
"""

from .rak_lib import RakLib
from .packet.unconnected_ping import UnconnectedPing
from .packet.unconnected_pong import UnconnectedPong



"""a Python software that gets basic information about an MCPE server

ServerData class

Copyright (c) 2016 w-gao
"""


class ServerData:

    # Query status
    SUCCESS = False

    # Request info
    PING_ID = -1
    SERVER_ID = -1

    GAME_ID = 'UNKNOWN'
    SERVER_NAME = 'UNKNOWN'
    GAME_PROTOCOL = -1
    GAME_VERSION = 'UNKNOWN'
    NUM_PLAYERS = -1
    MAX_PLAYERS = 1000

    HASH_CODE = -1
    MOTD = ''
    GAMEMODE = ''



"""a Python software that gets basic information about an MCPE server

Query class

Usage: Include the following code in your project
```
from py_mcpe_stats import Query

host = 'localhost'
port = 19132

q = Query(host, port)
server_data = q.query()
```

Copyright (c) 2016 w-gao
"""

import socket
from random import randint

from py_mcpe_stats import _server_data
from ._raknet import RakLib
from ._raknet import UnconnectedPing
from ._raknet import UnconnectedPong


class Query:

    def __init__(self, host, port=19132, timeout=5):
        self.host = host
        self.port = port
        self.timeout = timeout

        self.socket = None

    def query(self):
        # init socket
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.socket.settimeout(self.timeout)

            self.socket.connect((self.host, self.port))
        except socket.error as msg:
            print("Cannot connect to the server. Error: ", msg)
            return None

        # Returned data
        server_data = _server_data.ServerData()

        # get data from the server
        try:
            ping = UnconnectedPing(randint(1, 999999999))
            ping.encode()
            self.socket.send(ping.buffer)

            buff = self.socket.recv(65565)

            if buff[0] is RakLib.UNCONNECTED_PONG:
                pong = UnconnectedPong()
                pong.buffer = buff
                pong.decode()

                server_data.PING_ID = pong.ping_id
                server_data.SERVER_ID = pong.server_id

                # Trim
                info = pong.server_info.replace('\;', '').split(';')

                if len(info) > 6:
                    server_data.GAME_ID = info[0]
                    server_data.SERVER_NAME = info[1]
                    server_data.GAME_PROTOCOL = int(info[2])
                    server_data.GAME_VERSION = info[3]
                    server_data.NUM_PLAYERS = int(info[4])
                    server_data.MAX_PLAYERS = int(info[5])
                    server_data.HASH_CODE = info[6]
                    server_data.MOTD = info[7]
                    server_data.GAMEMODE = info[8]

                server_data.SUCCESS = True

        # Perhaps the server is offline
        except socket.error as msg:
            print('Failed to query. Error message: ', msg)

        # print('closing the socket')
        self.socket.close()
        return server_data



"""a Python software that gets basic information about an MCPE server

setup.py

Copyright (c) 2016 w-gao
"""

from setuptools import setup, find_packages

setup(name='py_mcpe_stats',
      version='0.1.3',
      description='a Python software that gets basic information about an MCPE server',
      long_description='a Python software that allows you to ping a Minecraft Pocket:Edition (MCPE) server for basic '
                       'information.',
      url='https://github.com/w-gao/py-mcpe-stats',
      author='w-gao',
      author_email='w-gao@users.noreply.github.com',
      license='MIT',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.5',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Networking'
      ],
      keywords='MCPE minecraft mcpeserver raknet',
      packages=find_packages('.')
      )



"""a Python software that gets basic information about an MCPE server

UnconnectedPong class

Copyright (c) 2016 w-gao
"""

from .packet import Packet


class UnconnectedPong(Packet):

    def __init__(self):
        super().__init__()

        self.ping_id = None
        self.server_id = None
        self.server_info = None

    def encode(self):
        pass

    def decode(self):

        self.ping_id = self.read_long()
        self.server_id = self.read_long()

        # Skip Magic
        self.offset += 16

        self.server_info = self.read_string()
